import sys
import pyvisa
from RSSpectrogram import RSSpectrogram
import time
import threading
from ScanDisplay import ScanDisplay
from SpectrogramData import SpectrogramData

class Instrument(SpectrogramData):
    def __init__(self, ESWdata:SpectrogramData):
        self.ESWdata = ESWdata
        self.rsspectrogram = RSSpectrogram(self.ESWdata.detectors)
        # Connect instrument
        self.rm = pyvisa.ResourceManager()
        self.rm.list_resources()
        self.ESW = self.rm.open_resource('TCPIP::xx.xxx.x.xxx::inst0::INSTR') 
        self.ESW.timeout = 500000
        print("IDN connected instrument: ", self.ESW.query("*IDN?"))

        self.ESW.query("*RST;*OPC?")
        self.ESW.query("INST:SEL REC;*OPC?")
        self.configScan()

    def configScan(self):
        """
        Configures the scan with the settings set in class Spectrogram_Data.
        :return: None
        """
        if self.ESWdata.settings["Mode"] == "Receiver":
            # Receiver Scan
            self.ESW.write("SCAN1:START " + str(self.ESWdata.settings["Start"]) + "MHz")
            self.ESW.write("SCAN1:STOP " + str(self.ESWdata.settings["Stop"]) + "MHz")
            self.ESW.write("SCAN1:TIME " + str(self.ESWdata.settings["Time"]))
            self.ESW.write("SCAN1:BAND:RES " + str(self.ESWdata.settings["Rbw"]) + "MHz")
            self.ESW.write("SCAN1:INP:TYPE INPUT1")
            self.ESW.write("SCAN1:INP:ATT 0dB")
        else:
            # Real-Time Spectrogram K55R
            self.ESW.write("INST:SEL RTSG")
            self.ESW.write("INP:PRES:COMP OFF")
            self.ESW.write("FREQ:STAR " + str(self.ESWdata.settings["Start"]) + "MHz")
            self.ESW.write("FREQ:STOP " + str(self.ESWdata.settings["Stop"]) + "MHz")
            self.ESW.write("INP:ATT:PROT OFF")
            self.ESW.write("INP:ATT 0dB")
            self.ESW.write("SWE:TIME " + str(self.ESWdata.settings["Time"]))

        self.ESW.write("FREQ:STAR " + str(self.ESWdata.settings["Start"]) + "MHz")
        self.ESW.write("FREQ:STOP " + str(self.ESWdata.settings["Stop"]) + "MHz")
        self.ESW.write("FREQ:TDOP FAST")
        self.ESW.write("CALC:UNIT:POW DBUV")

        # Set all detectors to on state and clear write
        for i in range(1,len(self.ESWdata.detectors)+1):
            self.ESW.write("DET"+str(i)+" "+ self.ESWdata.detectors[i-1])
            self.ESW.write("DISP:TRAC"+str(i)+" ON")
            self.ESW.write("DISP:TRAC"+str(i)+":MODE WRIT")

        # Set instrument to multimode to enable multiple spectrogram traces. If the instrument doesn't support multimode
        # e.g. with old firmware or other instrument, the program will exit
        self.ESW.write("CALC:SPEC:MMODe 1")
        if int(self.ESW.query("CALC:SPEC:MMODe?")) != 1:
            print("ESW does not support real-time output!")
            sys.exit(0)


    def runScan(self):
        # Start the scan in continuous mode
        self.ESW.write("INIT2:CONT ON")
        self.ESW.write("INIT2:IMM")

        last = 0
        read = 0
        first = -1

        while True:
            if last > read:
                if read < first and read > 1:
                    print("Requested frames out of instrument buffer!")
                    break
                # New frames available, read frames from instrument
                # Read frames from instrument with new SCPI command
                self.ESW.write("TRAC:SPEC:FDAT? " + str(read+1) + "," + str(last))
                re = self.ESW.read_raw()
                # Decode the raw data byte stream to spectrogram information
                self.ESWdata.data.extend(self.rsspectrogram.decodeSpectrogramBlock(re, self.ESWdata.settings["Time"]))

                # Check if the transmission is still real-time, if not, stop here
                if self.ESWdata.settings["CheckRealTime"] and not self.rsspectrogram.isRealTime():
                    print("Measurement time to low for ensuring real-time capturing and transmission!")
                    break
                read = last
                # If the current frame is the very first to receive, calculate the time difference between instrument and
                # PC clock to calibrate the graph delay calculation.
                if self.ESWdata.instrumentTimeDifference == 0:
                    self.ESWdata.instrumentTimeDifference = self.rsspectrogram.getInstrumentTimeDifference()
                last = self.rsspectrogram.availableIndex[1]
            else:
                # Ask instrument for availability of new frames
                finf = self.ESW.query("TRAC:SPEC:FINF?").split(",")
                first = int(finf[0])
                last = int(finf[1])
                #print("Available ", first, " to ", last)
                time.sleep(0.01)

    def stopScan(self):
        # Stop the instrument
        self.ESW.write("ABOR")
        self.ESW.query("*OPC?")


if __name__ == '__main__':
    ESWdata = SpectrogramData()
    ESWInstrument = Instrument(ESWdata)

    ESWGrapher = ScanDisplay(ESWdata)
    #ESWGrapher = SpectrogramDisplay(ESWdata)

    threading.Thread(target=ESWInstrument.runScan, args=()).start()
    threading.Thread(target=ESWGrapher.Update, args=()).start()
    ESWGrapher.canvas.app.run()






