import struct
import time
from collections import deque

class RSSpectrogram:
    def __init__(self, detectors):
        self.idx = 0
        self.firstFrame = True
        self.lastFrame = 1
        self.isrealtime = True
        self.availableIndex = [-1, -1]
        self.spectrogram = deque()
        self.detectors = detectors
        self.instrumentTimeDifference = 0

    def __readData(self, format, data):
        """
        Read date from byte stream
        :param format: Format defines the datatypes expected.
        :param data: Byte stream
        :return: List of decoded values from byte stream
        """
        length = struct.calcsize(format)
        unpack = struct.unpack_from(format, data, self.idx)
        self.idx += length
        return unpack

    def isRealTime(self):
        return self.isrealtime

    def getAvailableIndices(self):
        return self.availableIndex

    def getInstrumentTimeDifference(self):
        return self.instrumentTimeDifference

    def decodeSpectrogramBlock(self, data, measurementTime=0):
        """
        Decodes the spectrogram byte stream data to frame and trace information. To verify if the traces are captured in
        real-time the measurement time is needed as a parameter. Alternatively, the real-time detection can be disabled
        by setting the measurementTime to 0 (or leave it empty). This allows scanning over wider frequency spans, making
        the measurement non-real-time by the consecutive stepping over multiple segments.
        :param data: Byte information received as raw byte stream
        :param measurementTime: Measurement time as set in the receiver settings
        :return: All frames containing all traces each as a deque
        """
        self.spectrogram.clear()

        try:
            prefix = data[0:2].decode()
            x = int(prefix[1])
            # y = int(prefix[2:(2+x)])

            self.idx = x + 2

            header = self.__readData("<IddI", data)
            numberOfFrames = header[0]
            timeStampStart = header[1] + header[2] * 1e-9

            for f in range(1, numberOfFrames + 1):
                frame = self.__readData("<II", data)
                frameIndex = frame[0]
                numberOfTraces = frame[1]

                if frameIndex > self.lastFrame + 1 and measurementTime != 0:
                    # Real-time status lost as one frame was missed in capturing from the instrument
                    self.isrealtime = False

                # Empty dictionary to insert timestamp and trace data
                currentframe = {
                    "timestamp": 0,
                    **{d: None for d in self.detectors},
                }

                # Loop over all traces within the frame
                for t in range(1, numberOfTraces + 1):
                    trace = self.__readData("<IBddI", data)
                    timeStampStop = trace[2] + trace[3] * 1e-9

                    if measurementTime != 0 and not self.firstFrame and abs(timeStampStop - timeStampStart - measurementTime) > 1e-4:
                        # Detect violation of real-time by verifying the time difference between two consecutive frames
                        self.isrealtime = False

                    # Read the trace points as float32 numbers
                    currentframe[self.detectors[t - 1]] = self.__readData("<" + str(trace[4]) + "f", data)
                    #print("Read trace ", t, " of ", numberOfTraces, " from frame ", frameIndex, " with length ", str(
                    #    len(currentframe[self.detectors[t - 1]])), ". Start: ", timeStampStart, " Stop: ", timeStampStop)

                currentframe['timestamp'] = timeStampStop
                # Append the current frame to the spectrogram array of all captured frames
                self.spectrogram.append(currentframe)

                timeStampStart = timeStampStop
                self.lastFrame = frameIndex

                if self.firstFrame and frameIndex > 1:
                    # On the first two frames, the realtime check is disabled
                    self.instrumentTimeDifference = time.time() - timeStampStop
                    self.firstFrame = False

            # At the very end of each transmission is the available frames in the instrument buffer. A dedicated poll with
            # TRAC:SPEC:FINF? becomes obsolete as the information is also part of TRAC:SPEC:FDAT?
            self.availableIndex = self.__readData("<II", data)

        except Exception as error:
            print("Error at Spectrogram Decoding: ", error)
            self.isrealtime = False

        return self.spectrogram
