import math
import time
from array import array
from collections import deque
import numpy as np

class SpectrogramData:
    def __init__(self):
        self.export = False             # Export data in file while running
        self.exportSubsampling = 1      # Save every n-th element to file. A number greater than 1 will subsample trace data before saving.
        self.exportType = "CSV"         # BIN or CSV file format

        self.triggerEnable = False      # Enable trigger, False for freerun, True to trigger according to trigger function
        self.triggerLimit = 80          # Threshold in dBuV to trigger the data capturing
        self.triggerDuration = 10       # Duration of saved traces after trigger event in seconds
        self.graph_size = (1500,800)    # Size of the graph window

        """Choose a maximum total number of 4 detectors with a maximum of 2 CISPR detectors (CAV, CRMS, QPE) 
            and at maximum one of Average or RMS:
            AVER      Selects the Average detector
            CAV       Selects the CISPR Average detector
            CRMS      Selects the CISPR RMS detector
            POS       Selects the Max Peak detector
            QPE       Selects the Quasipeak detector
            RMS       Selects the RMS detector"""
        self.detectors = ["POS", "QPE", "CAV"]

        # Chose receiver settings. Start- and Stop Frequencies in MHz. Resolution bandwidth (Rbw) in MHz. Measurement
        # Time in seconds. Ref and Range define the reference level and the range in dB. Mode switches between Receiver
        # and RTSpectrum mode.
        self.settings = {"Start": 30, "Stop": 1000, "Time": 0.2, "Rbw": 0.12, "Ref": 90, "Range": 85, "Mode": "Receiver", "CheckRealTime": True}

        # End of settings, don't change anything from here
        self.lastTrigger = -1
        self.x_range = (self.settings["Start"], self.settings["Stop"])
        self.y_range = (self.settings["Ref"] - self.settings["Range"], self.settings["Ref"])
        self.points = math.ceil((self.settings["Stop"] - self.settings["Start"])/self.settings["Rbw"]*2)+1
        self.instrumentTimeDifference = 0
        self.dataDelay = 0
        self.data = deque()
        self.data_recorder = []

        self.exportQueue = deque()

    def trigger(self, frame):
        # Defines the trigger. In this example, the trigger is a simple horizontal threshold line that checks if the
        # first trace exceeds the threshold value at any frequency. Returns true if trigger is provoked.
        return np.max(frame[self.detectors[0]]) > self.triggerLimit


    def get_data_single(self):
        """
        This function fetches new data, enables triggering and exporting of trace data to a file.
        :return:   -1 no data available
                    1 if new data is available to display
                    2 if data is available but no trigger provoked
        """
        try:
            # Request one frame from the left side of the queue
            frame = self.data.popleft()

            # Check for triggering
            if self.triggerEnable & self.trigger(frame):
                # Trigger provoked, reset the trigger timer
                self.lastTrigger = frame['timestamp']
            if (not self.triggerEnable) | (int(self.lastTrigger + self.triggerDuration) > frame['timestamp']):
                if self.export:
                    self.exportQueue.append(frame)
                    # If at least 3 frames are in the export queue, write them into the file
                    if len(self.exportQueue) >= 3:
                        if self.exportType.lower() == "csv":
                            self.save_to_csv_file("storeData")
                        else:
                            self.save_to_binary_file("storeData")
            elif self.triggerEnable:
                if self.lastTrigger > 0:
                    self.lastTrigger = -1
                    self.data_recorder = 100*np.ones((len(frame[self.detectors[0]]) + 1, len(frame[self.detectors[0]])), dtype=np.float32)
                    self.data_recorder[0, :] = np.linspace(self.settings['Start'], self.settings['Stop'], len(frame[self.detectors[0]]))
                    return 1
                # Trigger is enabled but not provoked, no export and no display
                return 2

            # Calculate the delay between capturing on the instrument and display on the graph.
            # The time difference between timestamps is evaluated at the very first frame where we
            # assume to have zero delay between instrument and this program
            # data_recorder is needed for displaying the trace data using either the scan or spectrogram display
            self.dataDelay = max([time.time() - frame['timestamp'] - self.instrumentTimeDifference,0])

            self.data_recorder = np.zeros((len(frame[self.detectors[0]])+1, len(frame[self.detectors[0]])), dtype=np.float32)
            self.data_recorder[0,:] = np.linspace(self.settings['Start'], self.settings['Stop'], len(frame[self.detectors[0]]))

            for i in range(0,len(self.detectors)):
                self.data_recorder[i+1,:] = frame[self.detectors[i]]
            return 1
        except IndexError:
            # No frame available in the queue (yet)
            return -1

    def save_to_binary_file(self, filename):
        """
        Saves all items in the exportQueue into a binary file
        Binary representation takes the least amount of space on the disk while being the fastest to write. All frames
        are written consecutively with the timestamp in seconds as 8 byte float followed by all traces with 4 byte float
        values for each trace point. The number of written trace points depends on the exportSubsampling setting. There
        is no other information given in the byte stream nor any trace or frame separation.
        :param filename: filename without extension.
        """
        #ts = time.perf_counter()
        with open(filename+".bin", "ab") as file:
            while self.exportQueue:
                f = self.exportQueue.popleft()
                array('d',[f["timestamp"]]).tofile(file)
                for d in self.detectors:
                    array('f',f[d][1-1::self.exportSubsampling]).tofile(file)
        #t = time.perf_counter() - ts
        #print(f"Writing 3 Frames took: {t:.3f} sec")

    def save_to_csv_file(self, filename):
        """
        Exports the data to CSV. Please note that CSV export not only requires significantly more storage space, but also
        is the storing process much slower than saving to a binary file.
        For each frame, every trace is stored in a row with the following pattern:
        Timestamp in seconds, detector name, trace point 0, trace point 1, trace point 2, ...
        :param filename: filename without extension.
        """
        #ts = time.perf_counter()
        with open(filename+".csv", "a") as file:
            while self.exportQueue:
                f = self.exportQueue.popleft()
                for d in self.detectors:
                    file.write(str(f['timestamp']) + "," + d + ",")
                    np.savetxt(file, f[d][1-1::self.exportSubsampling], fmt="%f", delimiter=",", newline=",")
                    file.write("\n")
        #t = time.perf_counter() - ts
        #print(f"Writing 3 Frames took: {t:.3f} sec")